unit DFileStreams;

interface

uses
  Classes;

type
  _TDFileOpenMode = (
    dfBaseCreate,
    dfBaseAlways,
    dfBaseReadRight,
    dfBaseWriteRight,
    dfBaseDenyReadRight,
    dfBaseDenyWriteRight,
    dfBaseDoNotBuffer,
    dfBaseDelete
  );
  _TDFileOpenModes = set of _TDFileOpenMode;

  TDFileStream = class(THandleStream)
  public
    ChangeNotify: Boolean;

    constructor Create(const FileName: String; Mode: _TDFileOpenModes);
    destructor Destroy; override;
    function WriteStr(S: String): LongInt;
    function WriteByte(D: Byte): LongInt;
    function WriteWord(D: Word): LongInt;
    function WriteDWord(D: LongWord): LongInt;
    function ReadStr(MaxLen: LongInt = MAXINT): String;
    function ReadLStr: String;
    function ReadByte: Byte;
    function ReadWord: Word;
    function ReadDWord: LongWord;
    function ReadRIFFChunk(var Typ: LongWord): String;
  protected
    Path: String;
  end;

const
  dfCreate         : _TDFileOpenModes = [dfBaseCreate];
  dfAlways         : _TDFileOpenModes = [dfBaseAlways];
  dfOpenOrFail     : _TDFileOpenModes = [];
  dfOpenOrCreate   : _TDFileOpenModes = [dfBaseAlways];
  dfCreateOrFail   : _TDFileOpenModes = [dfBaseCreate];
  dfCreateOrReplace: _TDFileOpenModes = [dfBaseCreate, dfBaseAlways];

  dfRead     : _TDFileOpenModes = [dfBaseReadRight];
  dfWrite    : _TDFileOpenModes = [dfBaseWriteRight];
  dfReadWrite: _TDFileOpenModes = [dfBaseReadRight, dfBaseWriteRight];

  dfDenyRead : _TDFileOpenModes = [dfBaseDenyReadRight];
  dfDenyWrite: _TDFileOpenModes = [dfBaseDenyWriteRight];
  dfExclusive: _TDFileOpenModes = [dfBaseDenyReadRight, dfBaseDenyWriteRight];
  dfDenyNone : _TDFileOpenModes = [];

  dfNoBuffer : _TDFileOpenModes = [dfBaseDoNotBuffer];
  dfDelete   : _TDFileOpenModes = [dfBaseDelete];

implementation

uses
  SysUtils, Consts, Windows, ShlObj;

function ErrorText(ErrorCode: LongWord): String;
var
  WindowsError: PChar;

begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER + FORMAT_MESSAGE_FROM_SYSTEM, nil, ErrorCode, 0, PChar(@WindowsError), 0, nil);
  Result := WindowsError;
  LocalFree(Cardinal(WindowsError));
end;

constructor TDFileStream.Create(const FileName: String; Mode: _TDFileOpenModes);
var
  DesiredAccess,
  ShareMode,
  Creation,
  Flags: LongWord;
  NewHandle: THandle;
  Err: String;

begin
  Path := FileName;
  DesiredAccess := 0;
  if dfBaseReadRight in Mode then DesiredAccess := DesiredAccess or GENERIC_READ;
  if dfBaseWriteRight in Mode then DesiredAccess := DesiredAccess or GENERIC_WRITE;

  ShareMode := FILE_SHARE_READ or FILE_SHARE_WRITE;
  if dfBaseDenyReadRight in Mode then ShareMode := ShareMode and not FILE_SHARE_READ;
  if dfBaseDenyWriteRight in Mode then ShareMode := ShareMode and not FILE_SHARE_WRITE;
  if dfBaseDelete in Mode then ShareMode := ShareMode or FILE_SHARE_DELETE;

  if dfBaseCreate in Mode then begin
    if dfBaseAlways in Mode then
      Creation := CREATE_ALWAYS
    else
      Creation := CREATE_NEW
    ;
  end
  else begin
    if dfBaseAlways in Mode then
      Creation := OPEN_ALWAYS
    else
      Creation := OPEN_EXISTING
    ;
  end;

  Flags := FILE_ATTRIBUTE_NORMAL;
  if dfBaseDoNotBuffer in Mode then Flags := Flags or FILE_FLAG_NO_BUFFERING;
  if dfBaseDelete in Mode then Flags := Flags or FILE_FLAG_DELETE_ON_CLOSE;

  ChangeNotify := (Mode * [dfBaseWriteRight, dfBaseDelete]) <> [];

  NewHandle := CreateFile(
    PChar(FileName),
    DesiredAccess,
    ShareMode,
    nil,
    Creation,
    Flags,
    0
  );

  if NewHandle = 0 then NewHandle := INVALID_HANDLE_VALUE;
  if NewHandle = INVALID_HANDLE_VALUE then begin
    if dfBaseCreate in Mode then
      Err := 'Cannot create file ' + FileName
    else
      Err := 'Cannot open file ' + FileName
    ;
    Flags := GetLastError;
    Err := Err + ', error ' + IntToStr(Flags) + ': ' + ErrorText(Flags);
    raise EFOpenError.Create(Err);
  end
  else
    inherited Create(NewHandle)
  ;
end;

destructor TDFileStream.Destroy;
begin
  if THandle(Handle) = 0 then Exit;
  if THandle(Handle) = INVALID_HANDLE_VALUE then Exit;
  CloseHandle(Handle);
  
  if ChangeNotify then
    SHChangeNotify(SHCNE_UPDATEITEM, SHCNF_PATH, PChar(Path), nil)
  ;
end;

function TDFileStream.ReadByte: Byte;
var
  Left: Integer;

begin
  Left := Size - Position;
  if Left < 1 then raise EReadError.Create('Insufficient data to read a byte.');

  Left := Read(Result, 1);
  if Left <> 1 then raise EReadError.Create('Buffer underrun reading a byte.');
end;

function TDFileStream.ReadWord: Word;
var
  Left: Integer;

begin
  Left := Size - Position;
  if Left < 2 then raise EReadError.Create('Insufficient data to read a word.');

  Left := Read(Result, 2);
  if Left <> 2 then raise EReadError.Create('Buffer underrun reading a word.');
end;

function TDFileStream.ReadDWord: LongWord;
var
  Left: Integer;

begin
  Left := Size - Position;
  if Left < 4 then raise EReadError.Create('Insufficient data to read a D-word.');

  Left := Read(Result, 4);
  if Left <> 4 then raise EReadError.Create('Buffer underrun reading a D-word.');
end;

function TDFileStream.ReadStr(MaxLen: LongInt = MAXINT): String;
var
  Left: Integer;

begin
  Left := Size - Position;
  if MaxLen > Left then
    MaxLen := Left
  ;
  SetString(Result, nil, MaxLen);
  Left := Read(Result[1], MaxLen);
  if Left < MaxLen then
    SetLength(Result, Left)
  ;
end;

function TDFileStream.ReadLStr: String;
var
  L, Left: LongInt;

begin
  L := ReadDWord;
  Left := Size - Position;
  if L > Left then begin
    Seek(-4, soFromCurrent);
    raise EReadError.Create('Buffer underrun reading a lengthed string.');
  end;
  SetString(Result, nil, L);
  Read(Result[1], L);
end;

function TDFileStream.ReadRIFFChunk(var Typ: LongWord): String;
var
  Left: Integer;
  ClaimedSize, ActualSize: Integer;
  TypStr: String;

begin
  Left := Size - Position;
  if Left < 8 then raise EReadError.Create('Insufficient data to read a RIFF chunk header.');

  SetString(Result, nil, 8);
  ActualSize := Read(Result[1], 8);
  if ActualSize <> 8 then raise EReadError.Create('Buffer underrun reading RIFF chunk header.');
  Dec(Left, 8);
  TypStr := Copy(Result, 1, 4);
  Result := Copy(Result, 5, 4);
  Typ := PLongWord(PChar(TypStr))^;
  ClaimedSize := PLongWord(PChar(Result))^;

  if Left < ClaimedSize then raise EReadError.Create('Insufficient data to read content for RIFF chunk of type [' + TypStr + '].');

  SetString(Result, nil, ClaimedSize);
  ActualSize := Read(Result[1], ClaimedSize);
  if ActualSize <> ClaimedSize then raise EReadError.Create('Buffer underrun reading content for RIFF chunk of type [' + TypStr + '].');
end;

function TDFileStream.WriteStr(S: String): LongInt;
begin
  Result := Write(S[1], Length(S));
end;

function TDFileStream.WriteByte(D: Byte): LongInt;
begin
  Result := Write(D, SizeOf(D));
end;

function TDFileStream.WriteWord(D: Word): LongInt;
begin
  Result := Write(D, SizeOf(D));
end;

function TDFileStream.WriteDWord(D: LongWord): LongInt;
begin
  Result := Write(D, SizeOf(D));
end;

end.
